@@ -152,6 +152,9 @@ instance Monad m => MonadCatch (CatchT m) where
       Just e' -> runCatchT (c e')
       Nothing -> return (Left e)
     Right a -> return (Right a)
+-- | Note: This instance is only valid if the underlying monad has a single
+-- exit point!
+instance Monad m => MonadMask (CatchT m) where
   mask a = a id
   uninterruptibleMask a = a id
 
