@@ -15,6 +15,10 @@ import Data.Data (Data, Typeable)
 
 import Control.Monad.Trans.Identity (IdentityT(..))
 import Control.Monad.Reader (ReaderT(..))
+import Control.Monad.List (ListT(..))
+import Control.Monad.Trans.Maybe (MaybeT(..))
+import Control.Monad.Error (ErrorT(..))
+--import Control.Monad.Cont (ContT(..))
 import Test.Framework (Test, testGroup)
 import Test.Framework.Providers.QuickCheck2 (testProperty)
 import Test.QuickCheck (Property, once)
@@ -77,6 +81,11 @@ tests = testGroup "Control.Monad.Catch.Tests" $
         , MSpec "LazyRWS.RWST IO" $ \m -> io $ fmap tfst $ LazyRWS.evalRWST m () ()
         , MSpec "StrictRWS.RWST IO" $ \m -> io $ fmap tfst $ StrictRWS.evalRWST m () ()
 
+        , MSpec "ListT IO" $ \m -> io $ fmap (\[x] -> x) (runListT m)
+        , MSpec "MaybeT IO" $ \m -> io $ fmap (maybe undefined id) (runMaybeT m)
+        , MSpec "ErrorT IO" $ \m -> io $ fmap (either error id) (runErrorT m)
+        --, MSpec "ContT IO" $ \m -> io $ runContT m return
+
         , MSpec "CatchT Indentity" $ fromRight . runCatch
         ]
 
